package org.netkernel.layer0.bnf;

import org.netkernel.container.IMessages;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.urii.ActiveURIEncoding;
import org.netkernel.layer0.urii.URLQueryEncoding;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLUtils;
import org.w3c.dom.*;
import java.util.regex.*;
import java.util.*;

public class BNFGrammarFactory
{
	public static final String ARG_UNDEFINED_NAME="argName";
	public static final String ARG_UNDEFINED_VALUE="argValue";
	
	
	static final String ELE_GROUP="group";
	static final String ELE_CHOICE="choice";
	static final String ELE_INTERLEAVE="interleave";
	static final String ELE_OPTIONAL="optional";
	static final String ELE_REGEX="regex";
	static final String ATT_NAME="name";
	static final String ATT_MIN="min";
	static final String ATT_MAX="max";
	static final String ATT_TYPE="type";
	static final String ATT_ENC="encoding";

	static final BuiltIn[] BUILT_IN_REGEX = new BuiltIn[]
	{	new BuiltIn("uri","(([a-zA-Z][0-9a-zA-Z+\\-\\.]*:)?/{0,2}[0-9a-zA-Z;/?:@\\+&=$\\.\\-_!~*'\\(\\)%]+)?(#[0-9a-zA-Z;/?:@\\+&=$\\.\\-_!~*'\\(\\)%]+)?"),
		new BuiltIn("active-escaped-uri","(([a-zA-Z][0-9a-zA-Z+\\-\\.]*:)?/{0,2}[0-9a-zA-Z;/?:&=$\\.\\-_!~*'\\(\\)%]+)?(#[0-9a-zA-Z;/?:&=$\\.\\-_!~*'\\(\\)%]+)?"),
		new BuiltIn("active-escaped-uri-loose","[^\\+]*"),
		new BuiltIn("relative-path","([^/]+/)*[^/]+"),
		new BuiltIn("relative-directory-path","([^/]+/)*"),
		new BuiltIn("path-segment","[^/]+"),
		new BuiltIn("integer","[+-]?\\d+"),
		new BuiltIn("float","[+-]?\\d+(\\.\\d+)?"),
		new BuiltIn("nmtoken","[a-zA-Z][0-9a-zA-Z\\-\\._]*"),
		new BuiltIn("alphanum","[0-9a-zA-Z]+"),
		new BuiltIn("anything",".*"),
		new BuiltIn("nothing","[a&&[^a]]*"),
		//Leading (?i) makes regex match case insensitive
		new BuiltIn("iri","(?i)(?:[a-z](?:[-a-z0-9\\+\\.])*:(?:\\/\\/(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:])*@)?(?:\\[(?:(?:(?:[0-9a-f]{1,4}:){6}(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|::(?:[0-9a-f]{1,4}:){5}(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){4}(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:[0-9a-f]{1,4}:[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){3}(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:(?:[0-9a-f]{1,4}:){0,2}[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){2}(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:(?:[0-9a-f]{1,4}:){0,3}[0-9a-f]{1,4})?::[0-9a-f]{1,4}:(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:(?:[0-9a-f]{1,4}:){0,4}[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:(?:[0-9a-f]{1,4}:){0,5}[0-9a-f]{1,4})?::[0-9a-f]{1,4}|(?:(?:[0-9a-f]{1,4}:){0,6}[0-9a-f]{1,4})?::)|v[0-9a-f]+[-a-z0-9\\._~!\\$&'\\(\\)\\*\\+,;=:]+)\\]|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3}|(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=@])*)(?::[0-9]*)?(?:\\/(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@]))*)*|\\/(?:(?:(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@]))+)(?:\\/(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@]))*)*)?|(?:(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@]))+)(?:\\/(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@]))*)*|(?!(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@])))(?:\\?(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@])|[\\uE000-\\uF8FF\\uDC00\\uDB80-\\uDFFD\\uDBBF|\\uDC00\\uDBC0-\\uDFFD\\uDBFF\\/\\?])*)?(?:\\#(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@])|[\\/\\?])*)?|(?:\\/\\/(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:])*@)?(?:\\[(?:(?:(?:[0-9a-f]{1,4}:){6}(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|::(?:[0-9a-f]{1,4}:){5}(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){4}(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:[0-9a-f]{1,4}:[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){3}(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:(?:[0-9a-f]{1,4}:){0,2}[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){2}(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:(?:[0-9a-f]{1,4}:){0,3}[0-9a-f]{1,4})?::[0-9a-f]{1,4}:(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:(?:[0-9a-f]{1,4}:){0,4}[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3})|(?:(?:[0-9a-f]{1,4}:){0,5}[0-9a-f]{1,4})?::[0-9a-f]{1,4}|(?:(?:[0-9a-f]{1,4}:){0,6}[0-9a-f]{1,4})?::)|v[0-9a-f]+[-a-z0-9\\._~!\\$&'\\(\\)\\*\\+,;=:]+)\\]|(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])(?:\\.(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])){3}|(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=@])*)(?::[0-9]*)?(?:\\/(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@]))*)*|\\/(?:(?:(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@]))+)(?:\\/(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@]))*)*)?|(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=@])+)(?:\\/(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@]))*)*|(?!(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@])))(?:\\?(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@])|[\\uE000-\\uF8FF\\uDC00\\uDB80-\\uDFFD\\uDBBF|\\uDC00\\uDBC0-\\uDFFD\\uDBFF\\/\\?])*)?(?:\\#(?:(?:%[0-9a-f][0-9a-f]|[-a-z0-9\\._~\\u00A0-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF\\uDC00\\uD800-\\uDFFD\\uD83F\\uDC00\\uD840-\\uDFFD\\uD87F\\uDC00\\uD880-\\uDFFD\\uD8BF\\uDC00\\uD8C0-\\uDFFD\\uD8FF\\uDC00\\uD900-\\uDFFD\\uD93F\\uDC00\\uD940-\\uDFFD\\uD97F\\uDC00\\uD980-\\uDFFD\\uD9BF\\uDC00\\uD9C0-\\uDFFD\\uD9FF\\uDC00\\uDA00-\\uDFFD\\uDA3F\\uDC00\\uDA40-\\uDFFD\\uDA7F\\uDC00\\uDA80-\\uDFFD\\uDABF\\uDC00\\uDAC0-\\uDFFD\\uDAFF\\uDC00\\uDB00-\\uDFFD\\uDB3F\\uDC00\\uDB44-\\uDFFD\\uDB7F!\\$&'\\(\\)\\*\\+,;=:@])|[\\/\\?])*)?)")
		//new BuiltIn("","")
	};
	
	
	public static final BuiltIn BUILTIN_URI=BUILT_IN_REGEX[0];
	public static final BuiltIn BUILTIN_ACTIVE_ESCAPED_URI=BUILT_IN_REGEX[1];
	public static final BuiltIn BUILTIN_ACTIVE_ESCAPED_URI_SIMPLE=BUILT_IN_REGEX[2];
	public static final BuiltIn BUILTIN_RELATIVE_PATH=BUILT_IN_REGEX[3];
	public static final BuiltIn BUILTIN_RELATIVE_DIR=BUILT_IN_REGEX[4];
	public static final BuiltIn BUILTIN_PATH_SEGMENT=BUILT_IN_REGEX[5];
	public static final BuiltIn BUILTIN_INTEGER=BUILT_IN_REGEX[6];
	public static final BuiltIn BUILTIN_FLOAT=BUILT_IN_REGEX[7];
	public static final BuiltIn BUILTIN_NMTOKEN=BUILT_IN_REGEX[8];
	public static final BuiltIn BUILTIN_ALPHANUM=BUILT_IN_REGEX[9];
	public static final BuiltIn BUILTIN_ANYTHING=BUILT_IN_REGEX[10];
	public static final BuiltIn BUILTIN_NOTHING=BUILT_IN_REGEX[11];
	public static final BuiltIn BUILTIN_IRI=BUILT_IN_REGEX[12];
	
	public static final IPartEncoding NULL_ENCODING = new NullEncoder(); 
	
	public static IGrammarGroup NOTHING;
	static {
		NOTHING=BNFGrammarFactory.createGroup(null);
		BNFGrammarFactory.createRegex(NOTHING, BUILTIN_NOTHING);
	}
	
	public static class BuiltIn
	{	private String mName;
		private Pattern mRegex;
		private int mGroupCount;
		public BuiltIn(String aName, String aRegex)
		{	mName=aName;
			try
			{	mRegex=Pattern.compile(aRegex);
				mGroupCount=mRegex.matcher("").groupCount();
			}
			catch (Throwable e)
			{	e.printStackTrace();
			}
		}
		public String getName()
		{	return mName;
		}
		public Pattern getRegex()
		{	return mRegex;
		}
		public int getGroupCount()
		{	return mGroupCount;
		}
	}
	
	
		
	private static final PairList GRAMMAR_SCHEMA = new PairList(10);
	static
	{	GRAMMAR_SCHEMA.put("*[name!='group']", "root group expected");
		// GRAMMAR_SCHEMA.put("group[not(//regex or //choice)]", "no freedom");
		GRAMMAR_SCHEMA.put("group//*[name!='group' and name!='optional' and name!='choice' and name!='regex']", "invalid element, only group|optional|choice|regex valid");
		GRAMMAR_SCHEMA.put("group[//regex/*]", "regex cannot contain sub-elements");
		GRAMMAR_SCHEMA.put("group[//regex[@type and string-length(text())]]", "regex cannot have type and content");
		GRAMMAR_SCHEMA.put("group[//regex[not(@type) and not(string-length(text()))]]", "regex must have type or content");
		GRAMMAR_SCHEMA.put("group/attribute::*[name()!='encoding' and name()!='name' and name()!='min' and name()!='max' and name()!='desc']", "group attributes must be encoding,name,desc,min,max");
	}
	
	private static final PairList ENCODINGS = new PairList(2);
	static
	{	ENCODINGS.put("active", ActiveURIEncoding.INSTANCE);
		ENCODINGS.put("url-query", URLQueryEncoding.INSTANCE);
	}
	
	private static IPartEncoding getPartEncoding(String aName)
	{	IPartEncoding enc=NULL_ENCODING;
		if (aName!=null && aName.length()!=0)
		{	enc=(IPartEncoding)ENCODINGS.get(aName);
		}
		return enc;
	}
	
	
	private static Map<String,Pattern> sPatternCache = new HashMap();
	
	public static Pattern patternForString(String aPatternString)
	{	Pattern result=sPatternCache.get(aPatternString);
		if (result==null)
		{	result = Pattern.compile(aPatternString);
			sPatternCache.put(aPatternString, result);
		}
		return result;
	}
	
	
	
	public static IIdentifierGrammar parse(Element aRoot, IMessages aMessages) throws GrammarParseException
	{	PairList validationErrors = FastSchematron.validate(aRoot,GRAMMAR_SCHEMA);
		if (validationErrors.size()==0)
		{	IGrammarGroup rootGroup = (IGrammarGroup)recurseParse(aRoot,null,aMessages,true);
		
			rootGroup.close(aMessages);
			String startsWith=getStartsWith(rootGroup);
		
			return new GrammarImpl(rootGroup,startsWith);
		}
		else
		{	throw new GrammarParseException(validationErrors);
		}
	}
	
	public static IIdentifierGrammar createGrammar(IGrammarGroup aRootGroup, IMessages aMessages) throws GrammarParseException
	{	if (aRootGroup==null)
		{	throw new GrammarParseException(aMessages.format("MSG_BNF_NO_GRAMMAR"));
		}
		IGrammarGroup g=aRootGroup;
		while(g.getParent()!=null)
		{	g=g.getParent();
		}
		g.close(aMessages);
		
		String startsWith=getStartsWith(aRootGroup);
		
		g.close2();
		IIdentifierGrammar result = new GrammarImpl(g,startsWith);
		return result;
	}
	
	/** Find a static start string for the identifier if possible
	 * @param aGroup
	 * @return
	 */
	private  static String getStartsWith(IGrammarGroup aGroup)
	{	StringBuilder sb=new StringBuilder(128);
		recurseGroup(aGroup,sb);
		String result=sb.length()>0?sb.toString():null;
		return result;
	}
	
	private static boolean recurseGroup(IGrammarPart p, StringBuilder aSB)
	{	boolean active=true;
			
		if (p instanceof IGrammarGroup)
		{	
			IGrammarGroup g=(IGrammarGroup)p;
			if (g instanceof Choice || g instanceof Interleave)
			{	active=false;
			}
			else if (g instanceof RegularGroup && ((RegularGroup)g).getMin()==0)
			{	active=false;
			}
			else
			{	int pc=g.getPartCount();
				for (int i=0; i<pc && active; i++)
				{	active=recurseGroup( g.getPart(i), aSB);
				}
			}
		}
		else if (p instanceof StaticPart)
		{	aSB.append(((StaticPart)p).getContent());
		}
		else if (p instanceof RegexPart)
		{	
			String regex=((RegexPart) p).getRegex();
			int l=regex.length();
			int j=0;
			for (; j<l; j++)
			{	char c=regex.charAt(j);
				if (c=='[' || c=='.' || c=='(' || c=='\\' || c=='+')
				{	break;
				}
				if (c=='?' || c=='*')
				{	if (j>0) j--;
					break;
				}
				if (c=='|')
				{	j=0;
					break;
				}
			}
			if (j>0)
			{	aSB.append(regex.substring(0,j));
			}
			active=false;
		}
		else
		{	active=false;
		}
		return active;
	}
	
	public static IGrammarGroup createGroup(IGrammarGroup aParent)
	{	return createGroup(aParent,null,1,1,NULL_ENCODING);
	}
	public static IGrammarGroup createGroup(IGrammarGroup aParent, String aName, IPartEncoding aEncoding)
	{	if (aEncoding==null) aEncoding=NULL_ENCODING;
		return createGroup(aParent,aName,1,1,aEncoding);
	}
	public static IGrammarGroup createOptionalGroup(IGrammarGroup aParent, String aName, IPartEncoding aEncoding)
	{	return createGroup(aParent,aName,0,1,aEncoding);
	}
	public static IGrammarGroup createChoiceGroup(IGrammarGroup aParent)
	{	return createChoiceGroup(aParent,null,NULL_ENCODING);
	}
	public static IGrammarGroup createChoiceGroup(IGrammarGroup aParent, String aName, IPartEncoding aEncoding)
	{	if (aEncoding==null) aEncoding=NULL_ENCODING;
		IGrammarGroup result = new Choice(aParent,aName,aEncoding);
		if (aParent!=null)
		{	aParent.addPart(result);
		}
		return result;
	}
	public static IGrammarGroup createGroup(IGrammarGroup aParent, int aMin, int aMax)
	{	return createGroup(aParent,null,aMin,aMax,NULL_ENCODING);
	}
	public static IGrammarGroup createGroup(IGrammarGroup aParent, String aName, int aMin, int aMax, IPartEncoding aEncoding)
	{	if (aEncoding==null) aEncoding=NULL_ENCODING;
		GroupImpl result = new RegularGroup(aParent,aName,aMin,aMax,aEncoding);
		if (aParent!=null)
		{	aParent.addPart(result);
		}
		return result;
	}
	public static IGrammarGroup createInterleave(IGrammarGroup aParent)
	{	IGrammarGroup result = new Interleave(aParent);
		aParent.addPart(result);
		return result;
	}
	public static IGrammarPart createRegex(IGrammarGroup aParent, String aRegex)
	{	IGrammarPart result = new RegexPart(aRegex);
		aParent.addPart(result);
		return result;
	}
	public static IGrammarPart createRegex(IGrammarGroup aParent, BuiltIn aBuiltin)
	{	IGrammarPart result = new RegexPart(aBuiltin);
		aParent.addPart(result);
		return result;
	}
	public static IGrammarPart createText(IGrammarGroup aParent, String aText)
	{	IGrammarPart result = new StaticPart(aText);
		aParent.addPart(result);
		return result;
	}
	
	private static IGrammarGroup recurseParse(Element aLocation, IGrammarGroup aParent, IMessages aMessages, boolean aIsRoot) throws GrammarParseException
	{
		IGrammarGroup result;
		
		String tag=aLocation.getNodeName();
		
		String name=aLocation.getAttribute(ATT_NAME);
		if (name.length()==0)
		{	name=null;
		}

		String encString=aLocation.getAttribute(ATT_ENC);
		IPartEncoding enc = NULL_ENCODING;
		if (encString.length()>0)
		{	enc=getPartEncoding(encString);
			if (enc==null)
			{	throw new GrammarParseException(aLocation,aMessages.format("MSG_BNF_BAD_ENC",encString));
			}
		}
		
		if (tag.equals(ELE_CHOICE))
		{	result = createChoiceGroup(aParent,name,enc);
		}
		else if (tag.equals(ELE_OPTIONAL))
		{	result = createOptionalGroup(aParent, name,enc);
		}
		else if (tag.equals(ELE_INTERLEAVE))
		{	result = createInterleave(aParent);
		}
		else if (tag.equals(ELE_GROUP) || aIsRoot)
		{	int min=Utils.parseValency(aLocation.getAttribute(ATT_MIN),1);
			int max=Utils.parseValency(aLocation.getAttribute(ATT_MAX),1);
			if (min==-1 || max==-1 || min>max || min==Integer.MAX_VALUE)
			{	throw new GrammarParseException(aLocation,aMessages.format("MSG_BNF_BAD_QUANTIFIERS"));
			}
			result = createGroup(aParent, name, min, max, enc);
		}
		else
		{	throw new GrammarParseException(aLocation,aMessages.format("MSG_BNF_UNKNOWN_EL"));
		}
		
		for (Node n=aLocation.getFirstChild(); n!=null; n=n.getNextSibling() )
		{	
			if (n instanceof Element)
			{	Element child = (Element)n;
				tag=child.getNodeName();
				if (tag.equals(ELE_REGEX))
				{
					String builtin=child.getAttribute(ATT_TYPE);
					if (builtin.length()>0)
					{	BuiltIn found=null;
						for (int i=0; i<BUILT_IN_REGEX.length && found==null; i++)
						{	if (BUILT_IN_REGEX[i].getName().equals(builtin))
							{	found=BUILT_IN_REGEX[i];
							}
						}
						if (found!=null)
						{	createRegex(result, found);
						}
						else
						{	throw new GrammarParseException(aLocation,aMessages.format("MSG_BNF_BAD_BUILTIN",builtin));
						}
					}
					else
					{	String regex=XMLUtils.getText(child).trim();
						createRegex(result,regex);
					}					
				}
				else
				{	recurseParse(child,result,aMessages,false);
				}
			}
			else if (n instanceof Text)
			{	String text=n.getNodeValue().trim();
				if (text.length()>0)
				{	createText(result, text);
				}
			}
		}		
		
		return result;
	}
}
