package org.netkernel.layer0.bnf;

import java.io.IOException;
import java.io.StringWriter;
import org.netkernel.container.IMessages;
import org.netkernel.layer0.meta.GrammarException;
import org.netkernel.layer0.util.OrderedCheapMultiStringMap;
import org.netkernel.layer0.util.XMLUtils;

public class Choice extends GroupImpl
{
	public Choice(IGrammarGroup aParent, String aName, IPartEncoding aEncoding)
	{	super(aParent,aName,aEncoding);
	}
	
	public void writeXML(StringWriter aSW) throws IOException
	{	aSW.write("<choice");
		String groupName=getName();
		if (groupName!=null)
		{	aSW.write(" ");
			XMLUtils.writeAttribute(aSW,"name",groupName);
		}
		aSW.write(">");
		for (int i=0; i<getPartCount(); i++)
		{	IGrammarPart p = getPart(i);
			if (p instanceof IGrammarGroup)
			{	p.writeXML(aSW);
			}
			else
			{	aSW.write("<group>");
				p.writeXML(aSW);
				aSW.write("</group>");
			}
		}
		aSW.write("</choice>");
	}

	protected void buildRegex(StringWriter aSW)
	{	aSW.write('(');
		mGroups.putGroup(this);
		int count=getPartCount();
		for (int i=0; i<count; i++)
		{	getPart(i).writeRegex(aSW, mGroups );
			if (i<count-1)
			{	aSW.write('|');
			}
		}
		aSW.write(')');
	}
	
	public boolean matches(String[] aCapturedGroups, int aOffset)
	{	boolean result=false;
	
		for (int i=0; i<mPartsArray.length; i++)
		{	IGrammarPart p = mPartsArray[i];
			if (p instanceof IGrammarGroup)
			{	IGrammarGroup group = (IGrammarGroup)p;
				int j=group.getCapturedGroupIndexInParent();
				if (aCapturedGroups[j+aOffset]!=null)
				{	if (group.needsToMatchDeeper())
					{	result = group.matches(aCapturedGroups,j+aOffset);
					}
					else
					{	result=true;
					}
					break;
				}
			}
		}

		return result;	
	}

	public void parse(String[] aCapturedGroups, OrderedCheapMultiStringMap aResults, int aOffset)
	{
		String name=getName();
		String value=aCapturedGroups[aOffset];
		if (name!=null && value!=null)
		{	aResults.put(name, getEncoding().decode(value));
		}
		
		for (int i=0; i<mPartsArray.length; i++)
		{
			IGrammarPart p = mPartsArray[i];
			if (p instanceof IGrammarGroup)
			{	IGrammarGroup g=(IGrammarGroup)p;
				String whole=aCapturedGroups[aOffset];
				if (whole!=null && g.match(whole))
				{	g.parse(aCapturedGroups, aResults, aOffset+g.getCapturedGroupIndexInParent());
					return;
				}
			}
		}
	}

	public String constructIdentifier(OrderedCheapMultiStringMap aPartsRemaining, IMessages aMessages) throws GrammarException
	{
		
		String name=getName();
		if (name!=null)
		{	String result=(String)aPartsRemaining.remove(name);
			if (result!=null)
			{	return getEncoding().encode(result);
			}
			else
			{	throw new GrammarException(aMessages.format("MSG_BNF_MISSING", name));
			}
		}
		
		int count=getPartCount();
		String staticPart=null;
		
		//see if we have a subGroup match
		OrderedCheapMultiStringMap backupCopyOfPartsRemaining=new OrderedCheapMultiStringMap(aPartsRemaining);
		for (int i=0; i<count; i++)
		{	Object part = getPart(i);
			if (part instanceof IGrammarGroup)
			{	IGrammarGroup g=(IGrammarGroup)part;
				try
				{	String subGroup=g.constructIdentifier(aPartsRemaining,aMessages);
					if (aPartsRemaining.size()==backupCopyOfPartsRemaining.size())
					{	// nothing used when choosing this choice
						throw new GrammarException("");
					}
					return subGroup;
				}
				catch (GrammarException e)
				{	if (aPartsRemaining.size()!=backupCopyOfPartsRemaining.size())
					{	aPartsRemaining.restoreFrom(backupCopyOfPartsRemaining);
					}
				}
			}
			else if (part instanceof StaticPart)
			{	if (staticPart==null)
				{	StaticPart p = (StaticPart)part;
					staticPart=p.getContent();
				}
				else
				{	new GrammarException(aMessages.format("MSG_BNF_AMBIGUOUS"));
				}
			}
		}
		
		//otherwise look for a static part
		if (staticPart!=null)
		{	return staticPart;
		}
		else
		{	throw new GrammarException(aMessages.format("MSG_BNF_NO_CHOICE"));
		}
		
	}

}